from typing import Optional
from pydantic_settings import BaseSettings, SettingsConfigDict
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore


class Settings(BaseSettings):

    # ==================
    # Logging
    # ==================

    #: FastAPI debug mode
    debug: bool = False
    #: Default log level.  Choose from any of the standard Python log levels.
    loglevel: Literal['NOTSET', 'DEBUG', 'INFO', 'WARN', 'ERROR', 'CRITICAL'] = 'INFO'
    #: What format should we log in?  Valid values are ``json`` and ``text``
    log_type: Literal['json', 'text'] = 'text'

    # ==================
    # HTTP
    # ==================

    #: Use this as the title for the login form, to give a hint to the
    #: user as to what they're logging into
    auth_realm: str = 'Restricted'

    # ==================
    # Session
    # ==================

    #: The name of the cookie to set when a user authenticates
    cookie_name: str = 'nginxauth'
    #: The domain to use for our session cookie, if any.
    cookie_domain: Optional[str] = None
    #: The secret key to use for session cookies
    secret_key: str = 'SESSION_SECRET'
    #: The maximum age of a session cookie in seconds
    session_max_age: int = 0
    #: Session type: either ``redis`` or ``memory``
    session_backend: Literal['redis', 'memory'] = 'memory'
    #: If using the Redis session backend, the URL to connect to Redis on
    redis_url: str = 'redis://localhost'
    #: If using the Redis session backend, the prefix to use for session keys
    redis_prefix: str = "nginx_ldap_auth."

    # ==================
    # LDAP
    # ==================

    #: The URI via which to connect to LDAP
    ldap_uri: str
    #: The DN as which to bind to LDAP
    ldap_binddn: str
    #: The password to use when binding to LDAP when doing our searches
    ldap_password: str
    #: Whether to use TLS when connecting to LDAP
    ldap_starttls: bool = True
    #: Whether to disable LDAP referrals
    ldap_disable_referrals: bool = False
    #: The base DN under which to perform searches
    ldap_basedn: str = 'ou=users,dc=example,dc=com'
    #: The LDAP attribute to use as the username when searching for a user
    ldap_username_attribute: str = 'uid'
    #: The LDAP attribute to use as the full name when getting search results
    ldap_full_name_attribute: str = 'cn'
    #: The LDAP search filter to use when searching for a user.  This should
    #: be a valid LDAP search filter.  The search will be a SUBTREE search
    #: with the base DN of :py:attr:`ldap_basedn`.
    #:
    #: You may use these replacement fields in the filter:
    #:
    #: - ``{username_attribute}``: the value of :py:class:`Settings.ldap_username_attribute`
    #: - ``{username_full_name_attribute}``: the value of :py:class:`Settings.ldap_full_name_attribute`
    #:
    #: Use ``{username}`` in the search filter as the placeholder for the username
    #: supplied by the user from the login form.
    ldap_get_user_filter: str = '{username_attribute}={username}'
    #: The LDAP search filter to use to determine whether a user is authorized.  This
    #: should a valid LDAP search filter. If this is ``None``, all users who can successfully
    #: authenticate will be authorized.  If this is not ``None``, the search with this
    #: filter must return at least one result for the user to be authorized.
    #:
    #: You may use these replacement fields in the filter:
    #:
    #: - ``{username_attribute}``: the value of :py:attr:`ldap_username_attribute`
    #: - ``{username_full_name_attribute}``: the value of :py:attr:`ldap_full_name_attribute`
    #:
    #: Use ``{username}`` in the search filter as the placeholder for the username
    #: supplied by the user from the login form.
    ldap_authorization_filter: Optional[str] = None
    #: Number of seconds to wait for an LDAP connection to be established
    ldap_timeout: int = 15
    #: Min number of LDAP connections to keep in the pool
    ldap_min_pool_size: int = 1
    #: Max number of LDAP connections to keep in the pool
    ldap_max_pool_size: int = 30
    #: Recycle LDAP connections after this many seconds
    ldap_pool_connection_lifetime_seconds: int = 20

    # ==================
    # Sentry
    # ==================
    #: The sentry DSN to use for error reporting.  If this is ``None``, no
    #: error reporting will be done.
    sentry_url: Optional[str] = None

    model_config = SettingsConfigDict()



import os
import pprint

import click
import uvicorn

from ..settings import Settings

from .cli import cli

settings = Settings()


@cli.command('settings', short_help="Print our application settings.")
def print_settings():
    """
    Print our settings to stdout.  This should be the completely evaluated
    settings including those imported from any environment variable.
    """
    pp = pprint.PrettyPrinter(indent=2)
    pp.pprint(settings.model_dump())


@cli.command('start', short_help="Start the nginx_ldap_auth service.")
@click.option(
    '--host', '-h',
    default=lambda: os.environ.get("HOSTNAME", "0.0.0.0"),
    help="The host to listen on."
)
@click.option(
    '--port', '-p',
    default=lambda: os.environ.get("PORT", 8888),
    type=int,
    help="The port to listen on."
)
@click.option(
    '--reload/--no-reload', '-r',
    default=lambda: os.environ.get("RELOAD", False),
    type=bool,
    help="Reload the server on code changes."
)
@click.option(
    '--keyfile', '-k',
    default=lambda: os.environ.get('SSL_KEYFILE', '/certs/server.key'),
    type=click.Path(exists=True, dir_okay=False),
    help="The path to the SSL key file."
)
@click.option(
    '--certfile', '-c',
    default=lambda: os.environ.get('SSL_CERTFILE', '/certs/server.crt'),
    type=click.Path(exists=True, dir_okay=False),
    help="The path to the SSL certificate file."
)
@click.option(
    '--workers', '-w',
    default=lambda: os.environ.get('WORKERS', 1),
    type=int,
    help="The number of worker processes to spawn."
)
@click.option(
    '--env-file',
    default=None,
    type=click.Path(exists=True, dir_okay=False),
    help="The path to the environment file to load."
)
def start(**kwargs):
    """
    Start the nginx_ldap_auth service.
    """
    uvicorn_kwargs = {
        'host': kwargs['host'],
        'port': kwargs['port'],
        'reload': kwargs['reload'],
        'ssl_keyfile': kwargs['keyfile'],
        'ssl_certfile': kwargs['certfile'],
        'workers': kwargs['workers'],
        'ssl_version': 2,
    }
    if kwargs['env_file']:
        uvicorn_kwargs['env_file'] = kwargs['env_file']
    uvicorn.run("nginx_ldap_auth.app.main:app", **uvicorn_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os

from fastapi import FastAPI, Request, Response, status
from fastapi.responses import HTMLResponse, RedirectResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from starsessions import (
    InMemoryStore,
    SessionStore,
    load_session,
    get_session_handler,
)
from starsessions.stores.redis import RedisStore
import structlog

from nginx_ldap_auth import __version__
from nginx_ldap_auth.settings import Settings

from ..logging import get_logger

from .forms import LoginForm
from .middleware import SessionMiddleware
from .models import User

current_dir = os.path.dirname(os.path.abspath(__file__))
static_dir = os.path.join(current_dir, "static")
templates_dir = os.path.join(current_dir, "templates")


settings = Settings()

# --------------------------------------
# Session Store
# --------------------------------------

if settings.session_backend == "memory":
    store: SessionStore = InMemoryStore()
elif settings.session_backend == "redis":
    store = RedisStore(
        settings.redis_url,
        prefix=settings.redis_prefix,
        gc_ttl=settings.session_max_age
    )
    get_logger().info(
        "session.store",
        backend=settings.session_backend,
        server=settings.redis_url,
        max_age=settings.session_max_age
    )


# --------------------------------------
# The FastAPI app
# --------------------------------------

app = FastAPI(
    title="nginx_ldap_auth",
    debug=settings.debug,
    version=__version__
)
app.mount("/auth/static", StaticFiles(directory=static_dir), name="static")
app.add_middleware(
    SessionMiddleware,
    store=store,
    cookie_name=settings.cookie_name,
    lifetime=settings.session_max_age
)
templates = Jinja2Templates(directory=templates_dir)


# --------------------------------------
# Startup and Shutdown Events
# --------------------------------------

@app.on_event("startup")
async def startup() -> None:
    """
    Create the LDAP connection pool when we start up.
    """
    await User.objects.create_pool()


@app.on_event("shutdown")
async def shutdown() -> None:
    """
    Close the LDAP connection pool when we shut down.
    """
    await User.objects.cleanup()


# --------------------------------------
# Helper Functions
# --------------------------------------

async def kill_session(request: Request) -> None:
    """
    Kill the current session.

    This means empty the session object of all contents, and delete it from the
    backend.

    Args:
        request: The request object
    """
    for key in list(request.session.keys()):
        del request.session[key]
    await get_session_handler(request).destroy()


# --------------------------------------
# Views
# --------------------------------------

@app.get("/auth/login", response_class=HTMLResponse, name="login")
async def login(request: Request, service: str = "/"):
    """
    If the user is already logged in, redirect to the URI named by the ``service``,
    query paremeter, defaulting to ``/`` if that is not present.  Otherwise, render
    the login page.

    If the header ``X-Auth-Realm`` is set, use that as the title for the login
    page.  Otherwise use
    :py:attr:`nginx_ldap_auth.settings.Settings.auth_realm`.

    Args:
        request: The request object

    Keyword Args:
        service: redirect the user to this URL after successful login
    """
    auth_realm = request.headers.get("x-auth-realm", settings.auth_realm)
    _logger = get_logger(request)
    _logger.info("auth.login.start", target=service)
    await load_session(request)
    if request.session.get("username"):
        session_id = get_session_handler(request).session_id
        _logger.info(
            "auth.login.success.already_logged_in",
            username=request.session["username"],
            session_id=session_id,
            target=service,
        )
        return RedirectResponse(url=service)
    return templates.TemplateResponse(
        "login.html",
        {
            "request": request,
            "site_title": auth_realm,
            "service": service
        }
    )


@app.post("/auth/login", response_class=HTMLResponse, name="login_handler")
async def login_handler(request: Request):
    """
    Process our user's login request.  If authentication is successful, redirect
    to the value of the ``service`` hidden input field on our form.

    If authentication fails, display the login form again.

    If the header ``X-Auth-Realm`` is set, use that as the title for the login
    page.  Otherwise use
    :py:attr:`nginx_ldap_auth.settings.Settings.auth_realm`.

    Args:
        request: The request object
    """
    auth_realm = request.headers.get("x-auth-realm", settings.auth_realm)
    form = LoginForm(request)
    form.site_title = auth_realm
    await form.load_data()
    if await form.is_valid():
        await load_session(request)
        request.session["username"] = form.username
        return RedirectResponse(url=form.service, status_code=status.HTTP_302_FOUND)
    else:
        return templates.TemplateResponse("login.html", form.__dict__)


@app.get("/auth/logout", response_class=HTMLResponse, name="logout")
async def logout(request: Request):
    """
    Log the user out and redirect to the login page.

    Args:
        request: The request object
    """
    _logger = get_logger(request)
    await load_session(request)
    if username := request.session.get("username"):
        await kill_session(request)
        _logger.info("auth.logout", username=username)
    return RedirectResponse(url='/auth/login?service=/')


@app.get("/check")
async def check_auth(request: Request, response: Response):
    """
    Ensure the user is still authorized.  If the user is authorized, return
    200 OK, otherwise return 401 Unauthorized.

    The user is authorized if the cookie exists, the session the cookie refers
    to exists, and the ``username`` key in the settings is set.

    Additionally, the user must still exist in LDAP, and if
    :py:attr:`nginx_ldap_auth.settings.Settings.ldap_authorization_filter` is
    not ``None``, the user must also match the filter.

    Args:
        request: The request object
        response: The response object
    """
    if request.cookies.get(settings.cookie_name):
        await load_session(request)
        if request.session.get("username"):
            # We have a valid session
            if not await User.objects.get(request.session["username"]):
                # The user does not exist in LDAP; log them out
                await kill_session(request)
            if not await User.objects.is_authorized(request.session["username"]):
                # The user is not authorized
                await kill_session(request)
            return {}
        else:
            # Destroy the session because it is not valid
            await kill_session(request)
    # Force the user to authenticate
    response.headers["Cache-Control"] = "no-cache"
    response.status_code = status.HTTP_401_UNAUTHORIZED
    return {}



#!/usr/bin/env python
from setuptools import setup, find_packages

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()


setup(
    name="nginx-ldap-auth-service",
    version="2.0.3",
    description="A FastAPI app that authenticates users via LDAP and sets a cookie for nginx",
    author="Caltech IMSS ADS",
    author_email="imss-ads-staff@caltech.edu",
    url="https://github.com/caltechads/nginx-ldap-auth-service",
    packages=find_packages(exclude=["bin"]),
    install_requires=[
        "aiodogstatsd==0.16.0.post0",
        "fastapi>=0.95.0",
        "uvicorn[standard]==0.21.1",
        "bonsai==1.5.1",
        "pydantic-settings",
        "pydantic>=1.9.0",
        "structlog==23.1.0",
        "tabulate>=0.8.9",
        "click>=8.0",
        "sentry-sdk>=1.1.0",
        "jinja2>=3.0.1",
        "starsessions[redis]>=2.1.1",
        "python-multipart>=0.0.6",
        "watchfiles",
        "python-dotenv",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=['nginx', 'ldap', 'auth', 'fastapi', 'devops'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Development Status :: 3 - Alpha",
        "Framework :: FastAPI",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Developers",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: System :: Systems Administration :: Authentication/Directory :: LDAP",
    ],
    include_package_data=True,
    entry_points={
        'console_scripts': ['nginx-ldap-auth = nginx_ldap_auth.main:main']
    }
)
